# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
r"""

Class that constructs a set of emission-line templates, primarily for
use with :class:`~mangadap.proc.sasuke.Sasuke`.

Class usage examples
--------------------

To construct emission-line templates, you need a wavelength vector,
the instrumental resolution at which to construct the templates, and
an emission line database (see
:class:`mangadap.par.emissionlinedb.EmissionLineDB`). A simple
construction would be::
        
    # Imports
    import numpy
    from mangadap.par.emissionlinedb import EmissionLineDB
    from mangadap.proc.emissionlinetemplates import EmissionLineTemplates

    wave = numpy.logspace(*(numpy.log10([3600,10300]), 4563)
    sigma_inst = 30                     # Instrumental resolution in km/s

    tpl = EmissionLineTemplates(wave, sigma_inst, emldb=EmissionLineDB.from_key('ELPMILES'))

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""

import os
import time
import logging

from IPython import embed

import numpy
from scipy import interpolate

import astropy.constants

from ..util.sampling import spectrum_velocity_scale, spectral_coordinate_step
from ..util.log import log_output
from ..util import lineprofiles
from .spectralfitting import EmissionLineFit

# For debugging
from matplotlib import pyplot


class EmissionLineTemplates:
    r"""
    Construct a set of emission-line templates based on an emission-line
    database, primarily for use in the :class:`~mangadap.proc.sasuke.Sasuke`
    emission-line fitter.

    The templates are constructed based on the constraints provided by
    the emission-line database.  See
    :class:`~mangadap.par.emissionlinedb.EmissionLinePar` for the
    structure of each row in the database and an explanation for each of
    its columns.
    
    Only lines with ``action=f`` are included in any template.  The array
    :attr:`tpli` provides the index of the template that contains each
    line in the emission-line database.  Lines that are not assigned to
    any template --- either because they do not have ``action=f`` or their
    center lies outside the wavelength range in :attr:`wave` --- are
    given an index of -1.

    Only lines with ``mode=a`` (i.e., flux, velocity, and velocity
    dispersion are all tied) are included in the same template.

    Lines with tied velocities are assigned the same velocity component
    (:attr:`vgrp`) and lines with the tied velocity dispersions are
    assigned the same sigma component (:attr:`sgrp`).

    .. warning::

        The construction of templates for use with
        :class:`~mangadap.proc.sasuke.Sasuke` does
        *not* allow one to tie fluxes while leaving the velocities
        and/or velocity dispersions as independent.

    Args:
        wave (array-like):
            A single wavelength vector with the wavelengths for the
            template spectra.  The wavelengths are expected to be sample
            either linearly or geometrically (see :attr:`log`).
        sigma_inst (:obj:`float`, array-like):
            The single value or value as a function of wavelength for
            the instrumental dispersion (km/s) to use for the template
            construction.
        log (:obj:`bool`, optional):
            Flag that the wavelengths have been sampled geometrically.
        base (:obj:`float`, optional):
            Base for the geometric sampling.
        emldb (:class:`mangadap.par.emissionlinedb.EmissionLineDB`, optional):
            Emission-line database that is parsed to setup which lines
            to include in the same template because they are modeled as
            having the same velocity, velocity dispersion and flux
            ratio.  If not provided, no templates are constructed in
            instantiation; to build the templates using an existing
            instantiation, use :func:`build_templates`.
        flux_density (:obj:`bool`, optional):
            Return spectrum in units of flux density (flux per
            angstrom).  Default is to return the spectrum in units of
            flux per pixel.
        loggers (:obj:`list`, optional):
            List of `logging.Logger`_ objects to log progress; ignored
            if quiet=True.  Logging is done using
            :func:`mangadap.util.log.log_output`.  Default is no
            logging.  This can be reset in some methods.
        quiet (:obj:`bool`, optional):
            Suppress all terminal and logging output.

    Attributes:
        wave (`numpy.ndarray`_):
            Array with the wavelength (angstroms) of each pixel for all
            the constructed templates.  Shape is :math:`(N_{\rm pix},)`.
        sigma_inst (`scipy.interpolate.interp1d`_):
            The object used to interpolate the instrumental dispersion
            (km/s) at the rest wavelength of each spectral line.
        log (:obj:`bool`):
            Flag that the spectrum is sampled geometrically.
        base (:obj:`float`):
            Base for the geometric sampling.
        emldb (:class:`mangadap.par.emissionlinedb.EmissionLineDB`):
            Database with the emission-line parameters.  Shape is
            :math:`(N_{\rm pix},)`.
        ntpl (:obj:`int`):
            Total number of templates.
        flux (`numpy.ndarray`_):
            The template spectra.  Shape is :math:`(N_{\rm tpl},N_{\rm
            pix})`.
        tpli (`numpy.ndarray`_):
            The index of the template containing each emission line.
            Any emission-lines with ``tpli==-1`` means that the emission
            line was not included in any template, which should only
            occur for lines with ``action==i`` in the database.  Shape
            is :math:`(N_{\rm eml},)`.
        comp (`numpy.ndarray`_):
            The kinematic component assigned to each template.  Templates
            with the same component number are forced to have the same
            velocity and velocity dispersion by pPXF.  Shape is
            :math:`(N_{\rm tpl},)`.
        vgrp (`numpy.ndarray`_):
            The velocity group assigned to each template.  Templates in
            the same velocity group have their velocity parameters tied
            in pPXF, but the velocity dispersion parameters are
            independent.  Shape is :math:`(N_{\rm tpl},)`.
        sgrp (`numpy.ndarray`_):
            The velocity disperison (sigma) group assigned to each
            template.  Templates in the same sigma group have their
            velocity dispersion parameters tied in pPXF, but the
            velocity parameters are independent.  Shape is
            :math:`(N_{\rm tpl},)`.
        A_ineq (`numpy.ndarray`_):
            A matrix used to constrain the kinematic parameters between
            kinematic components. See the ``constr_kinem`` for pPXF (version
            7.0 and later). Shape is :math:`(N_{\rm constr}, N_{\rm comp})`;
            i.e., the number of constraints applied by the number of
            kinematic components.
        b_ineq (`numpy.ndarray`_):
            A vector used to constrain the kinematic parameters between
            kinematic components. See the ``constr_kinem`` for pPXF (version
            7.0 and later). Shape is :math:`(N_{\rm constr},)`; i.e., the
            number of constraints applied.
        eml_sigma_inst (`numpy.ndarray`_):
            The instrumental dispersion (km/s) at the rest wavelength
            of each emission line. This is mostly used to aid the
            velocity dispersion corrections determined by
            :class:`~mangadap.proc.sasuke.Sasuke`. Shape is
            :math:`(N_{\rm eml},)`.
        loggers (:obj:`list`):
            List of `logging.Logger`_ objects to log progress; ignored
            if quiet=True.  Logging is done using
            :func:`mangadap.util.log.log_output`.
        quiet (:obj:`bool`):
            Suppress all terminal and logging output.
    """
    def __init__(self, wave, sigma_inst, log=True, base=10, emldb=None, flux_density=True,
                 loggers=None, quiet=False):

        self.loggers=None
        self.quiet=None

        self.wave = numpy.asarray(wave)
        if len(self.wave.shape) != 1:
            raise ValueError('Provided wavelengths must be a single vector.')

        _sinst = numpy.full(self.wave.size, sigma_inst, dtype=float) \
                    if isinstance(sigma_inst, (int, float)) else numpy.asarray(sigma_inst)
        if _sinst.shape != self.wave.shape:
            raise ValueError('Provided sigma_inst must be a single number or a vector with the'
                             'same length as the wavelength vector.')
        self.sigma_inst = interpolate.interp1d(self.wave, _sinst, assume_sorted=True,
                                               bounds_error=False, fill_value=0)
        self.log = log
        self.base = base

        self.emldb = None           # Original database
        self.line_flux = None       # Line flux in the relevant template
        self.flux_density = None    # Templates are in flux per Angstrom, not flux per pixel
        self.ntpl = None            # Number of templates
        self.flux = None            # Template fluxes
        self.tpli = None            # Template associated with each emission line
        self.comp = None            # Kinematic component associated with each template
        self.vgrp = None            # Velocity group associated with each template
        self.sgrp = None            # Sigma group associated with each template
        self.A_ineq = None          # Linear inequality constraint matrix for the kinematics
        self.b_ineq = None          # Linear inequality constraint vector for the kinematics
        self.eml_sigma_inst = None  # Instrumental dispersion at the center of each line

        # If emission-line database is provided, use it to build the templates
        if emldb is not None:
            self.build_templates(emldb, flux_density=flux_density, loggers=loggers, quiet=quiet)

    def _parse_emission_line_database(self):
        r"""
        Parse the input emission-line database; see
        :class:`mangadap.par.emissionlinedb.EmissionLinePar`.

        Only lines with `action=f` are included in any template.  The
        array :attr:`tpli` provides the index of the template that
        contains each line in the emission-line database.  Lines that
        are not assigned to any template --- either because they do not
        have `action=f` or their center lies outside the wavelength
        range in :attr:`wave` --- are given an index of -1.

        Only lines with `mode=a` (i.e., tie flux, velocity, and velocity
        dispersion) are included in the same template.

        Lines with tied velocities are assigned the same velocity
        component (:attr:`vgrp`) and lines with the tied velocity
        dispersions are assigned the same sigma component
        (:attr:`sgrp`).

        .. warning::

            The construction of templates for use with
            :class:`~mangadap.proc.sasuke.Sasuke` does *not* allow
            one to tie fluxes while leaving the velocities and/or
            velocity dispersions as independent.

        This is an entirely internal procedure, taking no arguments and
        only assigning results to self.

        Raises:
            ValueError: Raised if the parsing of the database leaves
                some templates without an assigned component, velocity
                group, and/or sigma group.  This implies an error in the
                construction of the emission-line database, not the code
                itself.

        """
        # Get the list of lines to ignore
        ignore_line = (self.emldb['action'] != 'f') | (self.emldb['restwave'] < self.wave[0]) \
                        | (self.emldb['restwave'] > self.wave[-1])
        if numpy.any(ignore_line) and not self.quiet:
            log_output(self.loggers, 1, logging.INFO,
                       'Lines ignored because of action or wavelength limits: {0}'.format(
                       ', '.join(self.emldb['name'][ignore_line])))

        # Determine which parameters are tied by equality:
        tie_eq = numpy.array([p is not None and '=' in p 
                              for p in self.emldb['tie_par'].ravel()]).reshape(
                                  self.emldb['tie_par'].shape)
        tie_eq[ignore_line,:] = False

        # Determine which parameters re tied by inequality:
        tie_ineq = numpy.array([p is not None and '=' not in p
                                for p in self.emldb['tie_par'].ravel()]).reshape(
                                  self.emldb['tie_par'].shape) & numpy.logical_not(tie_eq)
        tie_ineq[ignore_line,:] = False

        # Ensure the problem is solvable.
        # TODO: I'm sure there's a more robust way to do this...
        if numpy.any(numpy.all(tie_eq | tie_ineq, axis=0)):
            raise ValueError('There must be at least one line that has an independent velocity '
                             'and another one that has an independent velocity dispsersion; '
                             'however, these do not have to be satisfied by a single line.')

        # Check that any tied line includes the index of the line it's tied to.
        untied = self.emldb['tie_index'] < 0
        indx = numpy.any(tie_eq | tie_ineq, axis=1) & untied
        if numpy.any(indx):
            raise ValueError('Index of reference line unspecified for {0}.'.format(
                             ', '.join(self.emldb['name'][indx])))

        # TODO: Need a check that the user hasn't input, e.g., '=0.3' for
        # anything other than the flux.
        if numpy.any([f is not None and '=' in f and len(f.strip('=')) > 0 
                        for f in self.emldb['tie_par'][:,1:].ravel()]):
            raise NotImplementedError('DAP cannot yet impose equality constraints for kinematics. '
                                      'Kinematics can only be forced to be identical (using \'=\')'
                                      ' or constrained to be within a fractional tolerance.')

        # Set the normalized line flux.  All default to unity.
        self.line_flux = numpy.ones(self.emldb.neml, dtype=float)
        # Specify those that are tied by equality
        self.line_flux[tie_eq[:,0]] = [float(f.strip('=')) 
                                        for f in self.emldb['tie_par'][tie_eq[:,0]][:,0]]

        # Currently cannot deal with lines that have tied fluxes but
        # independent kinematics
        if numpy.any(tie_eq[:,0] & numpy.logical_not(numpy.all(tie_eq[:,1:], axis=1))):
            raise NotImplementedError('DAP cannot currently accept lines with tied fluxes but '
                                      'independent kinematics.')

        # The total number of templates to construct is the number of lines in
        # the database minus the number of lines to ignore and the number of
        # lines where all parameters are tied by equality
        tied_all = numpy.all(tie_eq, axis=1)
        self.ntpl = self.emldb.size - numpy.sum(ignore_line) - numpy.sum(tied_all)

        # Initialize the components and tied kinematic groups
        self.comp = numpy.full(self.ntpl, -1, dtype=int)
        self.vgrp = numpy.full(self.ntpl, -1, dtype=int)
        self.sgrp = numpy.full(self.ntpl, -1, dtype=int)
        self.tpli = numpy.full(self.emldb.size, -1, dtype=int)

        # Reference lines are those that are:
        #   - *not* ignored (action != 'i'),
        #   - completely unconstrained (emldb['tie_index'] is indefined;
        #     i.e., less than 0), and/or
        #   - lines that are only tied by inequalities,
        # All reference lines are in separate templates, kinematic components,
        # velocity groups, and sigma groups
        ref_line = (untied | (numpy.logical_not(untied)
                              & numpy.logical_not(numpy.any(tie_eq[:,1:], axis=1)))) \
                    & numpy.logical_not(ignore_line)
        nref = numpy.sum(ref_line)

        self.tpli[ref_line] = numpy.arange(nref)
        self.comp[:nref] = numpy.arange(nref)
        self.vgrp[:nref] = numpy.arange(nref)
        self.sgrp[:nref] = numpy.arange(nref)

        finished = ref_line | ignore_line
        while numpy.sum(finished) != self.emldb.size:
            # Find the indices of lines that are tied to finished lines
            start_sum = numpy.sum(finished)
            for i in range(self.emldb.size):
                if finished[i]:
                    continue
                indx = numpy.where(self.emldb['index'] == self.emldb['tie_index'][i])[0]
                if not finished[indx]:
                    continue

                finished[i] = True

                # All line properties are tied such that the line will be part
                # of an existing template
                if numpy.all(tie_eq[i]):
                    self.tpli[i] = self.tpli[indx]
                # Both kinematic parameters are tied such that the line is part
                # of a different template but part of an existing kinematic
                # component
                elif numpy.all(tie_eq[i,1:]):
                    self.tpli[i] = numpy.amax(self.tpli)+1
                    self.comp[self.tpli[i]] = self.comp[self.tpli[indx]]
                    self.vgrp[self.tpli[i]] = self.vgrp[self.tpli[indx]]
                    self.sgrp[self.tpli[i]] = self.sgrp[self.tpli[indx]]
                # Line is part of a different template and kinematic component
                # with an untied sigma, but tied to an existing velocity group
                elif tie_eq[i,1]:
                    self.tpli[i] = numpy.amax(self.tpli)+1
                    self.comp[self.tpli[i]] = numpy.amax(self.comp)+1
                    self.sgrp[self.tpli[i]] = numpy.amax(self.sgrp)+1
                    self.vgrp[self.tpli[i]] = self.vgrp[self.tpli[indx]]
                # Line is part of a different template and kinematic component
                # with an untied velocity, but tied to an existing sigma group
                elif tie_eq[i,2]:
                    self.tpli[i] = numpy.amax(self.tpli)+1
                    self.comp[self.tpli[i]] = numpy.amax(self.comp)+1
                    self.vgrp[self.tpli[i]] = numpy.amax(self.vgrp)+1
                    self.sgrp[self.tpli[i]] = self.sgrp[self.tpli[indx]]

            # If the loop ends up with the same number of parsed lines
            # that it started with, there must be an error in the
            # construction of the input database.
            if start_sum == numpy.sum(finished):
                raise ValueError('Unable to parse the input database.  Check tying parameters.')

        # All templates must have an assigned component, velocity group and
        # dispersion group. Otherwise, something has gone wrong or the input
        # database hasn't been constructed correctly.
        if numpy.any(self.comp < 0) or numpy.any(self.vgrp < 0) or numpy.any(self.sgrp < 0):
            raise ValueError('Templates without an assigned component.  Check the input database.')

        # If there are no inequalities set in the database, we're done
        if not numpy.any(tie_ineq):
            self.A_ineq = None
            self.b_ineq = None
            return

        # Gut check
        assert not numpy.any(tie_eq & tie_ineq), \
                'CODING ERROR: Components are tied with both equality and inequality constraints'

        # The DAP currently doesn't deal with any template weight (flux)
        # inequalities:
        if numpy.any(tie_ineq[:,0]):
            raise NotImplementedError('DAP currently does not allow inequality constraints on '
                                      'line fluxes.')

        # Check that lines in the same component do not have different
        # constraints.
        # TODO: I'm not actually sure that it's possible for this to fault, but
        # it's here just in case.
        ncomp = numpy.amax(self.comp)+1
        # Component for each line
        compi = numpy.full(self.emldb.size, -1, dtype=int)
        indx = numpy.logical_not(ignore_line)
        compi[indx] = self.comp[self.tpli[indx]]
        for i in range(ncomp):
            indx = (compi == i) & tie_ineq[:,1]
            if numpy.any(indx) and len(numpy.unique(self.emldb['tie_par'][indx,1])) != 1:
                raise ValueError('Inequality constraints for velocities of lines '
                                 + '{0} ({1}) '.format(self.emldb['index'][indx],
                                                       self.emldb['name'][indx])
                                 + 'must be identical because they are in the same '
                                 + 'kinematic component.')
            indx = (compi == i) & tie_ineq[:,2]
            if numpy.any(indx) and len(numpy.unique(self.emldb['tie_par'][indx,2])) != 1:
                raise ValueError('Inequality constraints for dispersions of lines '
                                 + '{0} ({1}) '.format(self.emldb['index'][indx],
                                                       self.emldb['name'][indx])
                                 + 'must be identical because they are in the same '
                                 + 'kinematic component.')

        # The number of constraints is twice the number of specified
        # inequalities in the database
        indx = compi != -1
        nconstr = 2 * numpy.sum(tie_ineq[indx,1:])

        # Get the fractional constraint on each parameter
        tie_comp_par = numpy.zeros((ncomp, 2), dtype=float)
        # NOTE: This loop is necessary because of recursive tying of lines. For
        # example, in the OII doublet, one of the lines has its velocity tied
        # to the H-alpha line and the other has its velocity and velocity
        # dispersion tied to the first line. To make sure that the velocity
        # dispersion of both OII lines is within some inequality constraint of
        # the H-alpha dispersion, you need to pick the correct constraint. The
        # checks above should catch if this recursion leads to, e.g., different
        # constraints on the dispersion, so below I pick the correct constraint
        # by picking the maximum value. This works because the `=` constraint
        # is interpreted first as having a 0.0 fractional constraint.
        for i in range(ncomp):
            indx = compi == i
            _par = numpy.array([0.0 if f is None or len(f.strip('=')) == 0 else float(f.strip('='))
                                for f in self.emldb['tie_par'][indx,1:].ravel()]).reshape(
                                        numpy.sum(indx),2)
            tie_comp_par[i,:] = numpy.amax(_par, axis=0)

        # The definition of the fractional constraint is such that the value
        # MUST be larger than one. Determine if any of the values don't meet
        # this constraint and raise an exception if so.
        # TODO: Do this when reading the EmissionLineDB...
        if numpy.any(numpy.any((tie_comp_par < 0) 
                               | ((tie_comp_par > 0) & numpy.logical_not(tie_comp_par > 1)),
                               axis=1)):
            raise ValueError('Any defined inequality constraints must be >1!')

        # NOTE: For reference, this didn't work (compared to the explicit loop
        # above) because it was just assigning the most recent parameter with
        # the same component. Depending on the order of the line list, this
        # meant that some fractional constraints were set to 0, leading to a
        # singluar A_ineq array.
#        tie_comp_par[compi[indx],:] = numpy.array([0.0 if f is None or len(f.strip('=')) == 0 
#                                                 else float(f.strip('='))
#                                            for f in self.emldb['tie_par'][indx,1:].ravel()
#                                            ]).reshape(numpy.sum(indx),2)

        # Set a vector indicating the ineq connections between components
        tie_comp = numpy.full(ncomp, -1, dtype=int)
        indx_match = self.emldb.tie_index_match()
        indx = (indx_match >= 0) & numpy.any(tie_ineq[:,1:], axis=1) & (compi != -1)
        tie_comp[compi[indx]] = compi[indx_match[indx]]

        # Build the inequality for each component
        # NOTE: Inequality constraint is *always* defined with b_ineq=0 for
        # what the DAP can currently accommodate.
        self.b_ineq = numpy.zeros(nconstr, dtype=float)
        self.A_ineq = numpy.zeros((nconstr,2*ncomp), dtype=float)
        constr = 0
        for i in range(ncomp):
            if tie_comp[i] < 0:
                continue
            for j in range(2):
                if tie_comp_par[i,j] > 1:
                    self.A_ineq[constr,2*tie_comp[i]+j] = 1/tie_comp_par[i,j]
                    self.A_ineq[constr,2*i+j] = -1.
                    constr += 1
                    self.A_ineq[constr,2*tie_comp[i]+j] = -tie_comp_par[i,j]
                    self.A_ineq[constr,2*i+j] = 1.
                    constr += 1

    def build_templates(self, emldb, flux_density=True, profile='FFTGaussianLSF', loggers=None,
                        quiet=False):
        r"""
        Build the set of templates for a given emission-line database.
        The function uses the current values in :attr:`wave` and
        :attr:`sigma_inst`.  Any existing templates from a previous call
        to :func:`build_templates` or from the object instantiation will
        be overwritten using the provided emission-line database.

        See :func:`check_database` for the requirements of the provided
        emission-line database, and see
        :func:`_parse_emission_line_database` for how the database is
        interpretted when constructing the templates.

        The function constructs and returns the following attributes:
        :attr:`flux`, :attr:`comp`, :attr:`vgrp`, and :attr:`sgrp` 

        .. todo::

            - Warn the user if any line is undersampled; i.e., the FWHM
              of the line is less than 2.1 or :math:`\sigma < 0.9`.

            - Warn the user if any line grouped in the same template
              falls outside the spectral range.

        Args:
            emldb (:class:`mangadap.par.emissionlinedb.EmissionLineDB`):
                Emission-line database.
            flux_density (:obj:`bool`, optional):
                Return spectrum in units of flux density (flux per
                angstrom).  Default is to return the spectrum in units
                of flux per pixel.
            loggers (:obj:`list`, optional):
                List of `logging.Logger`_ objects to log progress;
                ignored if quiet=True.  Logging is done using
                :func:`mangadap.util.log.log_output`.
            quiet (:obj:`bool`, optional):
                Suppress all terminal and logging output.

        Returns:
            :obj:`tuple`: Returns 6 `numpy.ndarray`_ objects: (1) the set of
            templates with shape :math:`N_{\rm tpl}\times N_{\rm wave}`, (2)
            the kinematic component assignement for each template, (3) the
            velocity group associated with each template, (4) the sigma group
            associated with each template, (5) the A matrix used to impose
            constraints on the kinematics, and (6) the b vector used to
            impose constraints on the kinematics; for the latter two objects,
            see the ``constr_kinem`` keyword argument for pPXF.
        """
        #---------------------------------------------------------------
        # Initialize the reporting
        if loggers is not None:
            self.loggers = loggers
        self.quiet = quiet
        self.flux_density = flux_density

        #---------------------------------------------------------------
        # Check the database can be used with this class
        EmissionLineFit.check_emission_line_database(emldb, wave=self.wave)
        # Save a pointer to the database
        self.emldb = emldb
        # Parse the database for construction of the templates
        self._parse_emission_line_database()

        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO,
                       'Number of emission lines to fit: {0}'.format(numpy.sum(self.tpli>-1)))
            log_output(self.loggers, 1, logging.INFO,
                       'Number of emission-line templates: {0}'.format(len(self.comp)))
            log_output(self.loggers, 1, logging.INFO,
                       'Number of emission-line kinematic components: {0}'.format(
                                                                    numpy.amax(self.comp)+1))
            log_output(self.loggers, 1, logging.INFO,
                       'Number of emission-line kinematic constraints: {0}'.format(
                                                0 if self.b_ineq is None else self.b_ineq.size))
            log_output(self.loggers, 1, logging.INFO,
                       'Number of emission-line velocity groups: {0}'.format(
                                                                    numpy.amax(self.vgrp)+1))
            log_output(self.loggers, 1, logging.INFO,
                       'Number of emission-line sigma groups: {0}'.format(
                                                                    numpy.amax(self.sgrp)+1))

        # Get the instrumental dispersion at the center of each line
        self.eml_sigma_inst = self.sigma_inst(self.emldb['restwave'])

        # Convert from wavelengths to pixel coordinates
        _wave = numpy.log(self.wave)/numpy.log(self.base) if self.log else self.wave
        _dw = spectral_coordinate_step(self.wave, log=self.log, base=self.base)
        _restwave = numpy.log(self.emldb['restwave'])/numpy.log(self.base) if self.log \
                            else self.emldb['restwave']

        # Rest wavelength in pixel units
        _restwave = (_restwave - _wave[0])/_dw
        # Flux to pixel units; less accurate when spectrum is
        # logarithmically binned
        dl = self.emldb['restwave']*(numpy.power(self.base,_dw/2)-numpy.power(self.base,-_dw/2)) \
                        if self.log else _dw
        _flux = self.line_flux / dl if self.flux_density else self.line_flux
        # Dispersion in pixel units
        _sigma = self.eml_sigma_inst * self.emldb['restwave'] \
                        / astropy.constants.c.to('km/s').value / dl

        # Construct the templates
        _profile = eval('lineprofiles.'+profile)()  # Declare an instance of the desired profile
        pix = numpy.arange(self.wave.size)
        self.flux = numpy.zeros((self.ntpl,self.wave.size), dtype=float)
        for i in range(self.ntpl):
            # Find all the lines associated with this template:
            index = numpy.arange(self.emldb.size)[self.tpli == i]
            # Add each line to the template
            for j in index:
                # Use the first three moments of the line to set the
                # parameters
                p = _profile.parameters_from_moments(_flux[j], _restwave[j], _sigma[j])
                # Add the line to the flux in this template
                self.flux[i,:] += _profile(pix, p)

        return self.flux, self.comp, self.vgrp, self.sgrp, self.A_ineq, self.b_ineq







#!/usr/bin/env python3

import os
import time
import numpy

from matplotlib import pyplot, ticker, rc, colors

from astropy.io import fits
import astropy.constants

from mangadap.proc.util import growth_lim
from mangadap.util.bitmask import BitMask

#-----------------------------------------------------------------------------
def init_ax(fig, pos):
    ax = fig.add_axes(pos)#, facecolor='0.98')
    ax.minorticks_on()
    ax.tick_params(which='major', length=6) #, direction='in')
    ax.tick_params(which='minor', length=3) #, direction='in')
    return ax


def velocity_gradient(drpall, dapall):

    indx = (dapall['DAPDONE'] == 1) & (dapall['DAPTYPE'] == 'HYB10-GAU-MILESHC')
    indx &= (drpall['nsa_elpetro_ba'][dapall['DRPALLINDX']] > 0)

    ha_indx = indx & (dapall['HA_GVEL_LO_CLIP'] > -999) & (dapall['HA_GVEL_HI_CLIP'] > -999)
    ha_vdiff = dapall['HA_GVEL_HI_CLIP'][ha_indx] - dapall['HA_GVEL_LO_CLIP'][ha_indx]
    ha_sini = numpy.ma.sqrt(1-numpy.square(drpall['nsa_elpetro_ba'][dapall['DRPALLINDX'][ha_indx]]))

    st_indx = indx & (dapall['STELLAR_VEL_LO_CLIP'] > -999) & (dapall['STELLAR_VEL_HI_CLIP'] > -999)
    st_vdiff = dapall['STELLAR_VEL_HI_CLIP'][st_indx] - dapall['STELLAR_VEL_LO_CLIP'][st_indx]
    st_sini = numpy.ma.sqrt(1-numpy.square(drpall['nsa_elpetro_ba'][dapall['DRPALLINDX'][st_indx]]))

    mass_lim = [2e8, 5e11]
    velw_lim = [30,  4000]

    ha_sb_lim = [0.02, 8]
    st_sb_lim = [0.03, 1]

    font = { 'size' : 14 }
    rc('font', **font)

    w,h = pyplot.figaspect(1)
    fig = pyplot.figure(figsize=(1.5*w,1.5*h))

    ax = init_ax(fig, [0.3, 0.56, 0.4, 0.4])
    ax.set_xlim(mass_lim)
    ax.set_ylim(velw_lim)
    ax.tick_params(axis='x', which='both', direction='in')
    ax.set_xscale('log')
    ax.set_yscale('log')
    ax.xaxis.set_major_formatter(ticker.NullFormatter())

    x = numpy.linspace(6, 14, 100)
    y = x/4
    ax.plot(numpy.power(10,x), numpy.power(10,y-0.75), color='0.8', zorder=0)
    ax.plot(numpy.power(10,x), numpy.power(10,y), color='0.8', zorder=0)
    ax.plot(numpy.power(10,x), numpy.power(10,y+0.75), color='0.8', zorder=0)

#    y = 0.25 * x + 0.0
#    ax.plot(numpy.power(10,x), numpy.power(10,y), color='C3', zorder=1)

    sc = ax.scatter(drpall['nsa_sersic_mass'][dapall['DRPALLINDX'][ha_indx]],
               numpy.ma.divide(ha_vdiff, ha_sini),
               c=dapall['EMLINE_GSB_1RE'][ha_indx,18],
               lw=0, marker='.', s=10, zorder=2,
               norm=colors.LogNorm(vmin=ha_sb_lim[0], vmax=ha_sb_lim[1]),
               cmap='inferno_r')
    cax = fig.add_axes([0.54, 0.60, 0.15, 0.01])
    cb = pyplot.colorbar(sc, cax=cax, orientation='horizontal', ticks=[0.1,1],
                         format=ticker.ScalarFormatter())
    pyplot.setp(cb.ax.xaxis.get_ticklabels(), fontsize=10)
    cax.text(0.8, 2.5, r'$\langle I_{{\rm H}\alpha}\rangle_e$', ha='center', va='center',
             transform=cax.transAxes, fontsize=10)

    ax.text(-0.22, 0.5, r'$\Delta V_{\rm gas}$ [km/s]',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes,
            rotation='vertical')


    ax = init_ax(fig, [0.3, 0.15, 0.4, 0.4])
    ax.set_xlim(mass_lim)
    ax.set_ylim(velw_lim)
    ax.set_xscale('log')
    ax.set_yscale('log')

    x = numpy.linspace(6, 14, 100)
    y = x/4
    ax.plot(numpy.power(10,x), numpy.power(10,y-0.75), color='0.8', zorder=0)
    ax.plot(numpy.power(10,x), numpy.power(10,y), color='0.8', zorder=0)
    ax.plot(numpy.power(10,x), numpy.power(10,y+0.75), color='0.8', zorder=0)

    sc = ax.scatter(drpall['nsa_sersic_mass'][dapall['DRPALLINDX'][st_indx]],
                    numpy.ma.divide(st_vdiff, st_sini), c=dapall['SB_1RE'][st_indx],
                    lw=0, marker='.', s=10, zorder=2,
                    norm=colors.LogNorm(vmin=st_sb_lim[0], vmax=st_sb_lim[1]),
                    cmap='inferno_r')

    cax = fig.add_axes([0.31, 0.50, 0.11, 0.01])
    cb = pyplot.colorbar(sc, cax=cax, orientation='horizontal', ticks=[0.1,1],
                         format=ticker.ScalarFormatter())
    pyplot.setp(cb.ax.xaxis.get_ticklabels(), fontsize=10)
    cax.text(0.5, 2.5, r'$\langle I_g\rangle_e$', ha='center', va='center',
             transform=cax.transAxes, fontsize=10)

    ax.text(0.5, -0.18, r'$\mathcal{M}_\ast$ [$h^{-2} \mathcal{M}_\odot$]',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes)
    ax.text(-0.22, 0.5, r'$\Delta V_\ast$ [km/s]',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes,
            rotation='vertical')
    ax.text(0.55, 0.89, r'$\mathcal{M}_\ast\propto\ \Delta V^4$',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes,
            rotation=20, zorder=4)

    ofile = '../ms/figs/mass_vel.pdf'
    fig.canvas.print_figure(ofile, bbox_inches='tight')
    fig.clear()
    pyplot.close(fig)


#-----------------------------------------------------------------------------

if __name__ == '__main__':
    t = time.clock()

    dapall_file = '/Users/westfall/Work/MaNGA/testing/dr15/dapall-v2_4_3-2.2.1.fits'
    dapall = fits.open(dapall_file)['DAPALL'].data

    drpall_file = '/Users/westfall/Work/MaNGA/testing/dr15/drpall-v2_4_3.fits'
    drpall = fits.open(drpall_file)[1].data

    velocity_gradient(drpall, dapall)

    print('Elapsed time: {0} seconds'.format(time.clock() - t))





import os
from IPython import embed

import numpy
import astropy.constants

from mangadap.datacube import MaNGADataCube
from mangadap.proc.templatelibrary import TemplateLibrary #, available_template_libraries
from mangadap.util.sampling import spectrum_velocity_scale, spectral_coordinate_step
from mangadap.tests.util import requires_remote, remote_data_file, data_test_file


def test_read():
    for key in TemplateLibrary.supported_libraries:
        tpl = TemplateLibrary(key, match_resolution=False, velscale_ratio=1, spectral_step=1e-4,
                              log=True, hardcopy=False)

def test_mileshc():
    tpl = TemplateLibrary('MILESHC', match_resolution=False, velscale_ratio=4, spectral_step=1e-4,
                          log=True, hardcopy=False)

    # Shape
    assert tpl.ntpl == 42, 'Incorrect number of templates'
    assert tpl['FLUX'].data.shape == (42, 12634), 'Incorrect shape of the flux array'

    # Wavelength and velocity coordinates
    assert numpy.isclose(spectral_coordinate_step(tpl['WAVE'].data, log=True), 1e-4/4), \
                'Coordinate step is wrong'
    assert numpy.isclose(spectrum_velocity_scale(tpl['WAVE'].data),
                         astropy.constants.c.to('km/s').value*1e-4*numpy.log(10.)/4), \
                'Velocity step is wrong'

    # Mask
    nmasked = {}
    for k in tpl.bitmask.keys():
        nmasked[k] = numpy.sum(tpl.bitmask.flagged(tpl['MASK'].data, flag=k))

    assert numpy.sum(list(nmasked.values())) == 192, 'Masking has changed'
    assert nmasked['NO_DATA'] == 0, 'There should be no NO_DATA flags'

    # Flux values
    flux = numpy.ma.MaskedArray(tpl['FLUX'].data, mask=tpl['MASK'].data > 0)
    assert numpy.isclose(numpy.ma.mean(flux), 1.0), 'Templates should be normalized'


def test_write():
    directory, ofile = TemplateLibrary.default_paths('MILESHC', output_path=data_test_file())
    file_name = directory / ofile
    if file_name.exists():
        os.remove(str(file_name))
    tpl = TemplateLibrary('MILESHC', match_resolution=False, velscale_ratio=4, spectral_step=1e-4,
                          log=True, output_path=data_test_file(), hardcopy=True)
    assert file_name.exists(), 'File not written'
    os.remove(str(file_name))


@requires_remote
def test_match_resolution():
    cube = MaNGADataCube.from_plateifu(7815, 3702, directory_path=remote_data_file())
    tpl = TemplateLibrary('MILESHC', cube=cube, match_resolution=True, velscale_ratio=4,
                          hardcopy=False, output_path=remote_data_file())

    # Resolution should be virtually identical in unmasked regions
    indx = tpl['MASK'].data == 0
    assert numpy.std(tpl.sres(tpl['WAVE'].data[indx[0]]) - tpl['SPECRES'].data[0,indx[0]]) < 0.1, \
                'Spectral resolution difference is above tolerance.'

    # Check the file that would have been written has the expected path
    assert cube.directory_path == tpl.directory_path, 'Cube and TPL paths should match.'
    assert tpl.file_name().startswith(cube.output_root), 'TPL file should start with the cube root'




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
Implements a few base classes used during spectral fitting procedures.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import warnings

from IPython import embed

import numpy
from scipy.interpolate import interp1d
import astropy.constants

from ..util.bitmask import BitMask
from ..util.datatable import DataTable
from ..util.constants import DAPConstants
from ..util import lineprofiles
from ..util.filter import interpolate_masked_vector
from ..par.emissionlinedb import EmissionLineDB
from .bandpassfilter import emission_line_equivalent_width, passband_median

# For debugging
from matplotlib import pyplot


# BASE CLASS -----------------------------------------------------------
class SpectralFitting:
    """
    Base class for spectral fitting.
    """
    def __init__(self, fit_type, bitmask=None, par=None):
        self.fit_type = fit_type
        if bitmask is not None and not isinstance(bitmask, BitMask):
            raise TypeError('Input bit mask must have type BitMask.')
        self.bitmask = bitmask
        self.par = par


# ----------------------------------------------------------------------
# TODO: Minimize this to what any kinematics fit should *absolutely*
# provide. All the rest should go into separate DataTables specific to
# the fitting class.
class StellarKinematicsFitDataTable(DataTable):
    """
    Class defining the data table to hold the stellar kinematics fit.

    Table includes:

    .. include:: ../tables/stellarkinematicsfitdatatable.rst

    Args:
        ntpl (:obj:`int`):
            Number of templates used to model the stellar continuum.
        nadd (:obj:`int`):
            Number of coefficients in any additive polynomial
            included in the fit. Can be 0.
        nmult (:obj:`int`):
            Number of coefficients in any multiplicative polynomial
            included in the fit. Can be 0.
        nkin (:obj:`int`):
            Number of kinematic moments included in the fit. Note
            that the number of moments used as input guesses is
            *always* assumed to be 2.
        mask_dtype (:obj:`type`):
            The data type used for the maskbits (e.g., numpy.int16).
            Typically this would be set by
            :func:`mangadap.util.bitmask.BitMask.minimum_dtype`.
        shape (:obj:`int`, :obj:`tuple`, optional):
            The shape of the initial array. If None, the data array
            will not be instantiated; use :func:`init` to initialize
            the data array after instantiation.
    """
    def __init__(self, ntpl=1, nadd=0, nmult=0, nkin=2, mask_dtype=numpy.int16, shape=None):
        # NOTE: This should require python 3.7 to make sure that this
        # is an "ordered" dictionary.
        datamodel = dict(BINID=dict(typ=int, shape=None, descr='Bin ID number'),
                         BINID_INDEX=dict(typ=int, shape=None, descr='0-indexed number of bin'),
                         MASK=dict(typ=mask_dtype, shape=None, descr='Maskbit value'),
                         BEGPIX=dict(typ=int, shape=None,
                                     descr='Index of the first pixel included in the fit'),
                         ENDPIX=dict(typ=int, shape=None,
                                     descr='Index of the pixel just beyond the last pixel '
                                           'included in fit'),
                         NPIXTOT=dict(typ=int, shape=None, 
                                      descr='Total number of pixels in the spectrum to be fit.'),
                         NPIXFIT=dict(typ=int, shape=None,
                                      descr='Number of pixels used by the fit.'),
                         TPLWGT=dict(typ=float, shape=(ntpl,),
                                     descr='Optimal weight of each template.'),
                         TPLWGTERR=dict(typ=float, shape=(ntpl,),
                                        descr='Nominal error in the weight of each template.'),
                         USETPL=dict(typ=bool, shape=(ntpl,),
                                     descr='Flag that each template was included in the fit.'),
                         ADDCOEF=dict(typ=float, shape=(nadd,) if nadd > 1 else None,
                                      descr='Coefficients of the additive polynomial, if '
                                            'included.'),
                         MULTCOEF=dict(typ=float, shape=(nmult,) if nmult > 1 else None,
                                       descr='Coefficients of the multiplicative polynomial, if '
                                             'included.'),
                         KININP=dict(typ=float, shape=(2,),
                                     descr='Input guesses for the kinematics'),
                         KIN=dict(typ=float, shape=(nkin,),
                                     descr='Best-fitting stellar kinematics'),
                         KINERR=dict(typ=float, shape=(nkin,),
                                     descr='Errors in the best-fitting stellar kinematics'),
                         CHI2=dict(typ=float, shape=None,
                                   descr='Chi-square figure-of-merit for the fit'),
                         RCHI2=dict(typ=float, shape=None,
                                    descr='Reduced chi-square figure-of-merit for the fit'),
                         CHIGRW=dict(typ=float, shape=(5,),
                                    descr='Value of the error-normalized residuals at 0, 68%, '
                                          '95%, 99%, and 100% growth'),
                         RMS=dict(typ=float, shape=None,
                                  descr='Root-mean-square of the fit residuals.'),
                         RMSGRW=dict(typ=float, shape=(5,),
                                    descr='Value of absolute value of the fit residuals at 0, '
                                          '68%, 95%, 99%, and 100% growth'),
                         FRMS=dict(typ=float, shape=None,
                                   descr='Root-mean-square of the fractional residuals '
                                         '(i.e., residuals/model).'),
                         FRMSGRW=dict(typ=float, shape=(5,),
                                      descr='Value of absolute value of the fractional residuals '
                                            'at 0, 68%, 95%, 99%, 100% growth'),
                         SIGMACORR_SRES=dict(typ=float, shape=None,
                                             descr='Quadrature correction for the stellar '
                                                   'velocity dispersion determined by the mean '
                                                   'difference in spectral resolution between '
                                                   'galaxy and template data.'),
                         SIGMACORR_EMP=dict(typ=float, shape=None,
                                            descr='Quadrature correciton for the stellar '
                                                  'velocity dispersion determined by fitting the '
                                                  'optimal template to one resolution matched to '
                                                  'the galaxy data.'))

        keys = list(datamodel.keys())
        super(StellarKinematicsFitDataTable,
                self).__init__(keys, [datamodel[k]['typ'] for k in keys],
                               element_shapes=[datamodel[k]['shape'] for k in keys],
                               descr=[datamodel[k]['descr'] for k in keys],
                               shape=shape)


class StellarKinematicsFit(SpectralFitting):
    """
    Base class for fitting stellar kinematics.
    """
    def __init__(self, fit_method, bitmask, par=None):
        SpectralFitting.__init__(self, 'stellar_kinematics', bitmask=bitmask, par=par)
        self.fit_method = fit_method

    @staticmethod
    def init_datatable(ntpl, nadd, nmult, nkin, mask_dtype, shape=None):
        return StellarKinematicsFitDataTable(ntpl=ntpl, nadd=nadd, nmult=nmult, nkin=nkin,
                                             mask_dtype=mask_dtype, shape=shape)


# ----------------------------------------------------------------------
#class CompositionFit(SpectralFitting):
#    """
#    Base class for fitting the spectral composition.
#    """
#    def __init__(self, fit_method, bitmask, par=None):
#        SpectralFitting.__init__(self, 'composition', bitmask=bitmask, par=par)
#        self.fit_method = fit_method


# ----------------------------------------------------------------------
class EmissionLineFitDataTable(DataTable):
    """
    Primary data table with the results of the parameterized emission-line fit.

    Table includes:

    .. include:: ../tables/emissionlinefitdatatable.rst

    Args:
        neml (:obj:`int`):
            Number of emission lines being fit
        nkin (:obj:`int`):
            Number of kinematic parameters (e.g., 2 for V and sigma)
        mask_dtype (:obj:`type`):
            The data type used for the maskbits (e.g., numpy.int16).
            Typically this would be set by
            :func:`mangadap.util.bitmask.BitMask.minimum_dtype`.
        shape (:obj:`int`, :obj:`tuple`, optional):
            The shape of the initial array. If None, the data array
            will not be instantiated; use :func:`init` to initialize
            the data array after instantiation.
    """
    def __init__(self, neml=1, nkin=2, mask_dtype=numpy.int16, shape=None):
        # NOTE: This should require python 3.7 to make sure that this
        # is an "ordered" dictionary.
        datamodel = dict(BINID=dict(typ=int, shape=None, descr='Bin ID number'),
                         BINID_INDEX=dict(typ=int, shape=None, descr='0-indexed number of bin'),
                         FIT_INDEX=dict(typ=int, shape=(neml,),
                                        descr='The index in the fit database associated with '
                                              'each emission line.'),
                         MASK=dict(typ=mask_dtype, shape=(neml,),
                                   descr='Maskbit value for each emission line.'),
                         FLUX=dict(typ=float, shape=(neml,),
                                   descr='The best-fitting flux of the emission line.'),
                         FLUXERR=dict(typ=float, shape=(neml,),
                                      descr='The error in the best-fitting emission-line flux'),
                         KIN=dict(typ=float, shape=(neml,nkin),
                                  descr='The best-fitting kinematics in each emission line'),
                         KINERR=dict(typ=float, shape=(neml,nkin),
                                     descr='The error in the best-fitting emission-line '
                                           'kinematics'),
                         SIGMACORR=dict(typ=float, shape=(neml,),
                                        descr='Quadrature correction in the emission-line '
                                              'velocity dispersion'),
                         SIGMAINST=dict(typ=float, shape=(neml,),
                                        descr='Dispersion of the instrumental line-spread '
                                              'function at the location of each emission line.'),
                         SIGMATPL=dict(typ=float, shape=(neml,),
                                       descr='Dispersion of the instrumental line-spread function '
                                             'of the emission-line templates.'),
                         CONTAPLY=dict(typ=float, shape=(neml,),
                                       descr='The value of any additive polynomial included in '
                                             'the fit at the location of each emission line'),
                         CONTMPLY=dict(typ=float, shape=(neml,),
                                       descr='The value of any multiplicative polynomial included '
                                             'in the fit at the location of each emission line'),
                         CONTRFIT=dict(typ=float, shape=(neml,),
                                       descr='The value of any extinction curve included in the '
                                             'fit at the location of each emission line'),
                         LINE_PIXC=dict(typ=int, shape=(neml,),
                                        descr='The integer pixel nearest the center of each '
                                              'emission line.'),
                         AMP=dict(typ=float, shape=(neml,),
                                  descr='The best-fitting amplitude of the emission line.'),
                         ANR=dict(typ=float, shape=(neml,),
                                  descr='The amplitude-to-noise ratio defined as the model '
                                        'amplitude divided by the median noise in the two (blue '
                                        'and red) sidebands defined for the emission line.'),
                         LINE_NSTAT=dict(typ=int, shape=(neml,),
                                         descr='The number of pixels included in the fit metric '
                                               'calculations (LINE_RMS, LINE_FRMS, LINE_CHI2) '
                                               'near each emission line.'),
                         LINE_RMS=dict(typ=float, shape=(neml,),
                                       descr='The root-mean-square residual of the model fit '
                                             'near each emission line.'),
                         LINE_FRMS=dict(typ=float, shape=(neml,),
                                       descr='The root-mean-square of the fractional residuals '
                                             'of the model fit near each emission line.'),
                         LINE_CHI2=dict(typ=float, shape=(neml,),
                                       descr='The chi-square of the model fit near each emission '
                                             'line.'),
                         BMED=dict(typ=float, shape=(neml,),
                                   descr='The median flux in the blue sideband of each emission '
                                         'line'),
                         RMED=dict(typ=float, shape=(neml,),
                                   descr='The median flux in the red sideband of each emission '
                                         'line'),
                         EWCONT=dict(typ=float, shape=(neml,),
                                     descr='The continuum value interpolated at the '
                                           'emission-line center (in the observed frame) used for '
                                           'the equivalent width measurement.'),
                         EW=dict(typ=float, shape=(neml,),
                                 descr='The equivalent width of each emission line'),
                         EWERR=dict(typ=float, shape=(neml,),
                                    descr='The error in the equivalent width of each emission '
                                          'line'))

        keys = list(datamodel.keys())
        super(EmissionLineFitDataTable,
                self).__init__(keys, [datamodel[k]['typ'] for k in keys],
                               element_shapes=[datamodel[k]['shape'] for k in keys],
                               descr=[datamodel[k]['descr'] for k in keys],
                               shape=shape)


class EmissionLineFit(SpectralFitting):
    """
    Base class for fitting emission lines.
    """
    def __init__(self, fit_method, bitmask=None, par=None):
        SpectralFitting.__init__(self, 'emission_line', bitmask=bitmask, par=par)
        self.fit_method = fit_method

    @staticmethod
    def init_datatable(neml, nkin, mask_dtype, shape=None):
        return EmissionLineFitDataTable(neml=neml, nkin=nkin, mask_dtype=mask_dtype, shape=shape)

    @staticmethod
    def select_binned_spectra_to_fit(binned_spectra, minimum_snr=0.0, stellar_continuum=None,
                                     debug=False):
        """
        Select binned spectra for which to fit emission lines.

        .. todo::
            This could be based on the moment assessment of the
            emission-line S/N instead; for now just based on continuum
            S/N.

        Args:
            binned_spectra
                (:class:`mangadap.proc.spatiallybinnedspectra.SpatiallyBinnedSpectra`):
                Binned spectra to be fit.
            minimum_snr (float): The minimum S/N of the binned spectrum
                to fit; see
                :func:`mangadap.proc.spatiallybinnedspectra.SpatiallyBinnedSpectra.above_snr_limit`.
            stellar_continuum
                (:class:`mangadap.proc.stellarcontinuummodel.StellarContinuumModel`,
                optional): Stellar-continuum models that have been fit
                to the binned spectra, if available.  The current
                function will only return True for spectra that are both
                above the S/N limit and have good stellar-continuum
                models.

        Returns:
            numpy.ndarray: Boolean vector with the spectra in the
            binned_spectra object to fit.
        """

        bins_to_fit = binned_spectra.above_snr_limit(minimum_snr, debug=debug)
        if stellar_continuum is None:
            return bins_to_fit

        # Determine which spectra have a valid stellar continuum fit
        indx = numpy.invert(stellar_continuum.bitmask.flagged(
                                    stellar_continuum['PAR'].data['MASK'],
                                    flag=[ 'NO_FIT', 'INSUFFICIENT_DATA', 'FIT_FAILED']))
        with_good_continuum = numpy.zeros(binned_spectra.nbins, dtype=bool)
        with_good_continuum[stellar_continuum['PAR'].data['BINID_INDEX'][indx]] = True
        bins_to_fit &= with_good_continuum

        if debug:
            warnings.warn('DEBUG!!')
            indx = numpy.arange(len(bins_to_fit))[bins_to_fit]
#            numpy.random.shuffle(indx)
            bins_to_fit[indx[2:]] = False

        return bins_to_fit

    @staticmethod
    def select_spaxels_to_fit(binned_spectra, minimum_snr=None, bins_to_fit=None, debug=False):
        """
        Select spaxels for which to fit emission lines.
        """
        if not debug:
            good_spaxels = binned_spectra.check_fgoodpix()
            if minimum_snr is None:
                return good_spaxels
            return good_spaxels & binned_spectra.above_snr_limit(minimum_snr, original_spaxels=True)

        warnings.warn('DEBUG!!')

        uniq, indx = numpy.unique(binned_spectra['BINID'].data.ravel(), return_index=True)
        if uniq[0] == -1:
            uniq = uniq[1:]
            indx = indx[1:]
        spaxels_to_fit = numpy.zeros(numpy.prod(binned_spectra['BINID'].data.shape), dtype=bool)
        if bins_to_fit is None:
            spaxels_to_fit[indx] = True
        else:
            spaxels_to_fit[indx[bins_to_fit]] = True
        return spaxels_to_fit

#        spaxels_to_fit = binned_spectra.check_fgoodpix()
#        indx = numpy.arange(len(spaxels_to_fit))[spaxels_to_fit]
#        numpy.random.shuffle(indx)
#        spaxels_to_fit[indx[10:]] = False
#        return spaxels_to_fit

    @staticmethod
    def get_spectra_to_fit(binned_spectra, pixelmask=None, select=None, error=False,
                           original_spaxels=False):
        r"""
        Get the spectra to fit during the emission-line fitting.

        Args:
            binned_spectra
                (:class:`mangadap.proc.spatiallybinnedspectra.SpatiallBinnedSpectra`):
                Object with the spectra to fit.
            pixelmask
                (:class:`mangadap.util.pixelmask.SpectralPixelMask`,
                optional): Pixel mask to apply.
            select (`numpy.ndarray`, optional):
                Select specific spectra to return.  Must have the
                correct shape; cf. `original_spaxels`.
            error (:obj:`bool`, optional):
                Return :math:`1\sigma` errors instead of inverse
                variance.
            original_spaxels (:obj:`bool`, optional):
                Instead of the binned spectra, use the `cube` attribute
                of the `binned_spectra` object to return the original
                spaxels, corrected for Galactic extinction.

        Returns:
            Four objects are returned:
                - (1) The wavelength vector of the spectra,
                - (2) a masked numpy array with the flux data,
                - (3) a masked numpy array with the error data (returned
                  as either inverse variance or :math:`1\sigma`,
                - (4) and an array with the spectral resolution for each
                  spectrum, based on the internal binned spectra
                  parameters.
        """
        # Grab the spectra
        wave = binned_spectra['WAVE'].data.copy()

        if original_spaxels:
            flags = binned_spectra.cube.do_not_fit_flags()
            flux = binned_spectra.cube.copy_to_masked_array(flag=flags)
            ivar = binned_spectra.cube.copy_to_masked_array(attr='ivar', flag=flags)
            flux, ivar = binned_spectra.galext.apply(flux, ivar=ivar, deredden=True)
            sres = binned_spectra.cube.copy_to_array(attr='sres')
        else:
            flags = binned_spectra.do_not_fit_flags()
            flux = binned_spectra.copy_to_masked_array(flag=flags)
            ivar = binned_spectra.copy_to_masked_array(ext='IVAR', flag=flags)
            sres = binned_spectra.copy_to_array(ext='SPECRES')
            sres = numpy.apply_along_axis(interpolate_masked_vector, 1,
                                    numpy.ma.MaskedArray(sres, mask=numpy.invert(sres > 0)))
        nspec = flux.shape[0]

        # Convert inverse variance to error
        if error:
            ivar = numpy.ma.power(ivar, -0.5)
            flux[numpy.ma.getmaskarray(ivar)] = numpy.ma.masked

        # Mask any pixels in the pixel mask
        if pixelmask is not None:
            indx = pixelmask.boolean(binned_spectra['WAVE'].data, nspec=nspec)
            flux[indx] = numpy.ma.masked
            ivar[indx] = numpy.ma.masked
        
        _select = numpy.ones(nspec, dtype=bool) if select is None else select
        return wave, flux[_select,:], ivar[_select,:], sres[_select,:]

    @staticmethod
    def check_and_prep_input(wave, flux, ivar=None, mask=None, sres=None, continuum=None,
                             redshift=None, dispersion=None, default_dispersion=100.0):
        """
        Check the input used for emission-line measurements.

        inverse variance is converted to 1-sigma error

        mask must be a boolean array.

        sres can be a single vector, but will be returned as an array
        with a size that matches flux.

        output all converted to masked arrays with at least two dimensions

        """
        # Check the input wavelength and flux shapes
        if len(wave.shape) != 1:
            raise ValueError('Input wavelengths must be a single vector.')

        # Check the mask shape
        if mask is not None and mask.shape != flux.shape:
            raise ValueError('Input mask must have the same shape as the flux array.')

        # Convert the input arrays to masked arrays if they aren't
        # already, and compare the array shapes
        _flux = numpy.ma.atleast_2d(flux if isinstance(flux, numpy.ma.MaskedArray) \
                                            else numpy.ma.MaskedArray(flux, mask=mask))
        if len(wave) != _flux.shape[1]:
            raise ValueError('Wavelength vector does not match shape of the flux array.')
        expected_shape = _flux.shape
        nspec = expected_shape[0]

        if ivar is None:
            _err = None
        else:
            _ivar = numpy.ma.atleast_2d(ivar if isinstance(ivar, numpy.ma.MaskedArray) \
                                                else numpy.ma.MaskedArray(ivar, mask=mask))
            if _ivar.shape != expected_shape:
                raise ValueError('Input ivar array must be the same shape as the flux array.')
            _err = numpy.ma.sqrt(1.0 /_ivar)

        if sres is None:
            _sres = None
        else:
            _sres = numpy.array([sres]*nspec) if len(sres.shape) == 1 else sres.copy()
            if _sres.shape != expected_shape:
                raise ValueError('Input sres array must match the flux array, either as a single '
                                 'vector as an 2D array.')

        if continuum is None:
            _continuum = None
        else:
            _continuum = numpy.ma.atleast_2d(continuum \
                                    if isinstance(continuum, numpy.ma.MaskedArray) \
                                    else numpy.ma.MaskedArray(continuum, mask=mask))
            if _continuum.shape != expected_shape:
                raise ValueError('Input continuum array must be the same shape as the flux array.')

        # Check the input redshifts and dispersions
        _redshift = numpy.zeros(nspec, dtype=float) if redshift is None else redshift
        if len(_redshift) != nspec:
            raise ValueError('Must provide one redshift per input spectrum.')

        _dispersion = numpy.full(nspec, default_dispersion, dtype=float) \
                            if dispersion is None else dispersion
        if len(_dispersion) != nspec:
            raise ValueError('Must provide one dispersion per input spectrum.')

        # Return all arrays (even if None)
        return _flux, _err, _sres, _continuum, _redshift, _dispersion

    @staticmethod
    def subtract_continuum(flux, continuum):
        """
        Subtract the continuum.  Does not check that shapes match.
        Returns the continuum subtracted flux and a boolean array
        setting where the continuum is not defined.
        """
        # Allow continuum to be None
        if continuum is None:
            return flux, None

        # Get where the continuum is masked but the spectra are not
        no_continuum = numpy.invert(numpy.ma.getmaskarray(flux)) & numpy.ma.getmaskarray(continuum)

        # Subtract the continuum (ensure output is a masked array)
        continuum_subtracted_flux = numpy.ma.subtract(flux, continuum)

        # Unmask regions where only the continuum is masked.
        continuum_subtracted_flux.mask[no_continuum] = False

        return continuum_subtracted_flux, no_continuum

    @staticmethod
    def instrumental_dispersion(wave, sres, restwave, cz):
        """
        Determine the instrumental dispersion for a set of rest
        wavelengths and velocities.

        Args:

            wave (numpy.ndarray): Vector with the wavelengths of the
                spectrum.

            sres (numpy.ndarray): Vector with the spectral resolution as
                a function of wavelength.

            restwave (float, numpy.ndarray): Rest wavelengths for a set
                of measured lines.

            cz (float, numpy.ndarray): Redshifts (in km/s) of each or
                all lines.

        Returns:
            numpy.ndarray : The instrumental dispersions for each
                provided line.
        """
        # Check input
        if len(wave.shape) != 1:
            raise ValueError('Input wavelength must be a 1D vector.')
        if wave.shape != sres.shape:
            raise ValueError('Input wavelength and resolution vectors must have the same shape.')
        
        nwave = wave.size
        _restwave = numpy.atleast_1d(restwave)
        nline = _restwave.size
        _cz = numpy.atleast_1d(cz)
        if _cz.size not in [ 1, nline ]:
            raise ValueError('Must provide single redshift or one redshift per line.')
        _cz = numpy.full(nline, cz, dtype=float) if _cz.size == 1 else _cz

        interpolator = interp1d(wave, sres, fill_value='extrapolate', assume_sorted=True)
        c = astropy.constants.c.to('km/s').value

        sinst = c / interpolator((cz/c + 1.0) * restwave)/DAPConstants.sig2fwhm

#        pyplot.plot(wave, c/sres/DAPConstants.sig2fwhm, lw=2)
#        pyplot.plot(wave, 2.5*astropy.constants.c.to('km/s').value/wave/DAPConstants.sig2fwhm,
#                    color='k', lw=1)
#        pyplot.scatter((cz/c + 1.0) * restwave, sinst, color='C1', marker='.', lw=0, s=100)
#        pyplot.show()

        return c / interpolator((cz/c + 1.0) * restwave)/DAPConstants.sig2fwhm

    @staticmethod
    def check_emission_line_database(emldb, wave=None, check_par=True):
        r"""
        Check the emission-line database.  Modes are checked by
        :class:`mangadap.par.emissionlinedb.EmissionLinePar`, and the
        indices are checked to be unique by
        :class:`mangadap.par.emissionlinedb.EmissionLineDB`.

            - The type of the object must be
              :class:`mangadap.par.emissionlinedb.EmissionLineDB`
            - The provided profile type of each line must be a defined
              class.
            - At least one line must have ``mode='f'``
            - All tied lines must be tied to a line with a correctly
              specified index.
            - Warnings will be provided for any line with a centroid
              that falls outside of the provided wavelength range.
            - The database must provide at least one valid line.

        Args:
            emldb (:class:`mangadap.par.emissionlinedb.EmissionLineDB`):
                Emission-line database.
            wave (array-like):
                Wavelength vector.
            check_par (:obj:`bool`, optional):
                Validate the provided parameters.

        Raises:
            TypeError:
                Raised if the provided object is not an instance of
                :class:`mangadap.par.emissionlinedb.EmissionLineDB`.
            ValueError:
                Raised if any line has a mode of `x` or if the database
                does not provide a valid definition for any templates.
            NameError:
                Raised if a defined profile type is not known.

        """

        # Check the object type
        if not isinstance(emldb, EmissionLineDB):
            raise TypeError('Emission lines must be defined using an EmissionLineDB object.')

        # Check the profile type
        unique_profiles = numpy.unique(emldb['profile'])
        for u in unique_profiles:
            try:
                eval('lineprofiles.'+u)
            except NameError as e:
                raise NameError('Profile type {0} not defined in'
                                'mangadap.util.lineprofiles!'.format(u))

        # There must be one primary line
        if not numpy.any([m[0] == 'f' for m in emldb['mode']]):
            raise ValueError('At least one line in the database must have mode=f.')

        # Check that there are lines to fit
        lines_to_fit = emldb['action'] == 'f'
        if numpy.sum(lines_to_fit) == 0:
            raise ValueError('No lines to fit in the database!')
        if wave is not None:
            _wave = numpy.asarray(wave)
            if len(_wave.shape) != 1:
                raise ValueError('Provided wavelengths must be a single vector.')
            lines_in_range = numpy.array([rw > _wave[0] and rw < _wave[-1] 
                                                for rw in emldb['restwave']]) 
            if numpy.sum(lines_to_fit & lines_in_range) == 0:
                raise ValueError('No lines to fit in the provided spectral range!')

        # Check that the tied line indices exist in the database
        for m in emldb['mode']:
            if m[0] == 'f':
                continue
            tied_index = int(m[1:])
            if numpy.sum(emldb['index'] == tied_index) == 0:
                raise ValueError('No line with index={0} to tie to!'.format(tied_index))

        # Only check the provided parameters if requested
        if not check_par:
            return

        # Check the provided parameters, fix flags, and bounds
        for i in range(emldb.size):
            profile = eval('lineprofiles.'+emldb['profile'][i])
            npar = len(profile.param_names)
            if emldb['par'][i].size != npar*emldb['ncomp'][i]:
                raise ValueError('Provided {0} parameters, but expected {1}.'.format(
                                  emldb['par'][i].size, npar*emldb['ncomp'][i]))
            if emldb['fix'][i].size != npar*emldb['ncomp'][i]:
                raise ValueError('Provided {0} fix flags, but expected {1}.'.format(
                                  emldb['fix'][i].size, npar*emldb['ncomp'][i]))
            if numpy.any([f not in [0, 1] for f in emldb['fix'][i] ]):
                warnings.warn('Fix values should only be 0 or 1; non-zero values interpreted as 1.')
            if emldb['lobnd'][i].size != npar*emldb['ncomp'][i]:
                raise ValueError('Provided {0} lower bounds, but expected {1}.'.format(
                                  emldb['lobnd'][i].size, npar*emldb['ncomp'][i]))
            if emldb['hibnd'][i].size != npar*emldb['ncomp'][i]:
                raise ValueError('Provided {0} upper bounds, but expected {1}.'.format(
                                  emldb['hibnd'][i].size, npar*emldb['ncomp'][i]))
            if emldb['log_bnd'][i].size != npar*emldb['ncomp'][i]:
                raise ValueError('Provided {0} log boundaries designations, but expected '
                                 '{1}.'.format(emldb['log_bnd'][i].size, npar*emldb['ncomp'][i]))

    @staticmethod
    def check_emission_line_database(emldb, wave=None):
        r"""
        Check the emission-line database.  Modes are checked by
        :class:`mangadap.par.emissionlinedb.EmissionLinePar`, and the
        indices are checked to be unique by
        :class:`mangadap.par.emissionlinedb.EmissionLineDB`.

            - The type of the object must be
              :class:`mangadap.par.emissionlinedb.EmissionLineDB`
            - At least one line must be fit independently
            - All tied lines must be tied to a line with a correctly
              specified index.
            - The database must provide at least one valid line to fit.

        Args:
            emldb (:class:`mangadap.par.emissionlinedb.EmissionLineDB`):
                Emission-line database.
            wave (array-like):
                Wavelength vector.

        Raises:
            TypeError:
                Raised if the provided object is not an instance of
                :class:`mangadap.par.emissionlinedb.EmissionLineDB`.
            ValueError:
                Raised if there are no independently fitted lines, if there
                are no lines to fit, if there are no valid lines within the
                provided wavelength range, if the wavelength range is not
                provided as a 1D vector, or if the index to tie lines is not
                valid. (Warning: The latter does not catch lines with valid
                indices but where the tied line falls outside of the provided
                wavelength range.)
        """
        # Check the object type
        if not isinstance(emldb, EmissionLineDB):
            raise TypeError('Emission lines must be defined using an EmissionLineDB object.')

        # There must be one primary line
        if numpy.all(emldb['tie_index'] > 0):
            raise ValueError('At least one line in the database must be independent of all '
                             'other lines.  I.e., the tied index must be None.')

        # Check that there are lines to fit
        lines_to_fit = emldb['action'] == 'f'
        if numpy.sum(lines_to_fit) == 0:
            raise ValueError('No lines to fit in the database!')
        if wave is not None:
            _wave = numpy.asarray(wave)
            if len(_wave.shape) != 1:
                raise ValueError('Provided wavelengths must be a single vector.')
            lines_in_range = numpy.array([rw > _wave[0] and rw < _wave[-1] 
                                                for rw in emldb['restwave']]) 
            if numpy.sum(lines_to_fit & lines_in_range) == 0:
                raise ValueError('No lines to fit in the provided spectral range!')

        # Check that the tied line indices exist in the database
        for i in emldb['tie_index']:
            if i <= 0:
                continue
            if numpy.sum(emldb['index'] == i) == 0:
                raise ValueError('No line with index={0} to tie to!'.format(tied_index))

    @staticmethod
    def measure_equivalent_width(wave, flux, emission_lines, model_eml_par, mask=None,
                                 redshift=None, bitmask=None, checkdb=True):
        """
        The flux array is expected to have size Nspec x Nwave.

        Provided previous emission-line fits, this function adds the
        equivalent width measurements to the output database.

        Errors currently *do not* include the errors in the continuum
        measurement; only the provided error in the flux.

        Raises:
            ValueError:
                Raised if the length of the spectra, errors, or mask
                does not match the length of the wavelength array;
                raised if the wavelength, redshift, or dispersion arrays
                are not 1D vectors; and raised if the number of
                redshifts or dispersions is not a single value or the
                same as the number of input spectra.
        """

        # Check the input emission-line database
        if checkdb:
            EmissionLineFit.check_emission_line_database(emission_lines)

        nspec = flux.shape[0]
        nbands = len(emission_lines)
        if redshift is None:
            # If the redshift is NOT provided, use the fitted velocity
            # for each emission line.  Replace masked measurements with
            # mean (unmasked) redshift of that spectrum, or of all
            # spectra.
            _redshift = model_eml_par['KIN'][:,:,0]/astropy.constants.c.to('km/s').value
            _redshift = numpy.ma.MaskedArray(_redshift,
                                             mask=bitmask.flagged(model_eml_par['MASK']))
            mean_redshift = numpy.ma.mean(_redshift, axis=1).filled(numpy.ma.mean(_redshift))
            mean_redshift = numpy.array([mean_redshift]*nbands).T
            _redshift[_redshift.mask] = mean_redshift[_redshift.mask]
            _redshift = numpy.asarray(_redshift)
        else:
            # Use provided data and check its shape
            _redshift = redshift
            if _redshift.ndim == 1:
                if len(_redshift) != nspec:
                    raise ValueError('Must provide at least one redshift per input spectrum.')
                _redshift = numpy.array([_redshift]*nbands).T
            if _redshift.ndim == 2 and _redshift.shape != (nspec,nbands):
                raise ValueError('Provided redshift array does not match spectra and bands.')

        # Calculate the wavelength at which to measure the continuum,
        # matching what is done by
        # :class:`mangadap.proc.emissionlineMoments.EmissionLineMoments`
        line_center = (1+_redshift)*emission_lines['restwave'][None,:]

        # Compute the equivalent widths.  The checking done by
        # EmissionLineFit.check_and_prep_input is *identical* to what is
        # done within emission_line_equivalent_width()
        model_eml_par['BMED'], model_eml_par['RMED'], pos, model_eml_par['EWCONT'], \
                model_eml_par['EW'], model_eml_par['EWERR'] \
                        = emission_line_equivalent_width(wave, flux, emission_lines['blueside'],
                                                         emission_lines['redside'], line_center,
                                                         model_eml_par['FLUX'], mask=mask,
                                                         redshift=_redshift,
                                                         line_flux_err=model_eml_par['FLUXERR'])

        # Flag non-positive measurements
        if bitmask is not None:
            model_eml_par['MASK'][numpy.invert(pos)] \
                    = bitmask.turn_on(model_eml_par['MASK'][numpy.invert(pos)],
                                      'NON_POSITIVE_CONTINUUM')

    @staticmethod
    def line_metrics(emission_lines, wave, flux, ferr, model_flux, model_eml_par,
                     mask=None, model_mask=None, bitmask=None, window=15, fill_redshift=False):
        r"""
        Calculate fit-quality metrics near each emission line.

        .. todo::
            - Allow window to be defined in angstroms?

        Args:
            emission_lines (:class:`mangadap.par.emissionlinedb.EmissionLineDB`):
                Emission-line database use during the fit.
            wave (`numpy.ndarray`):
                Wavelength vector for object spectra.  Shape is
                :math:`(N_{\rm pix},)`.
            flux (`numpy.ndarray`):
                Object spectra that have been fit.  Can be provided as a
                `numpy.ma.MaskedArray`_.  Shape is :math:`(N_{\rm
                spec},N_{\rm pix})`.
            ferr (`numpy.ndarray`)
                :math:`1\sigma` errors in the object spectra.  Can be
                provided as a `numpy.ma.MaskedArray`_.  Shape is
                :math:`(N_{\rm spec},N_{\rm pix})`.
            model_flux (`numpy.ndarray`):
                Best-fitting model spectra. Can be provided as a
                `numpy.ma.MaskedArray`_.  Shape is (:math:`N_{\rm
                spec},N_{\rm pix}`).
            model_eml_par (:class:`EmissionLineFitDataTable`):
                A numpy record array with data type given by
                :func:`_per_emission_line_dtype`.  Uses ``FLUX``,
                ``KIN``, and ``MASK``; and assigns results to ``LINE_*``
                columns.
            mask (`numpy.ndarray`, optional):
                A mask for the object spectra that have been fit.  Added
                to mask attribute of `flux` if it is a
                `numpy.ma.MaskedArray`_.
            model_mask (`numpy.ndarray`, optional):
                A *boolean* numpy array with the mask for the model
                spectra.  Added to mask attribute of `model_flux` if it
                is a `numpy.ma.MaskedArray`_.
            bitmask (:class:`mangadap.util.bitmask.BitMask`, optional):
                The `BitMask` object used to interpret the MASK column
                in the `model_eml_par` object.  If None, the MASK column
                is ignored.
            window (:obj:`int`, optional):
                The width of the window used to compute the metrics
                around each line in number of pixels.
            fill_redshift (:obj:`bool`, optional):
                Fill any masked velocity measurement to the masked
                median of the velocities for the unmasked lines in the
                same spectrum when constructing the redshifted bands.
                If False, the A/N measurement is masked.

        Returns:
            `numpy.recarray`: Return the input `model_eml_par` after
            filling the LINE_PIXC, AMP, ANR, LINE_NSTAT, LINE_CHI2,
            LINE_RMS, and LINE_FRMS columns.

        Raises:
            ValueError:
                Raised if various checks of the input array sizes are
                incorrect.
        """
        # Check shapes and set masks for all arrays
        _wave = numpy.atleast_1d(wave)
        if _wave.ndim > 1:
            raise ValueError('Wavelength must be a 1D vector.')

        _flux = flux if isinstance(flux, numpy.ma.MaskedArray) else numpy.ma.MaskedArray(flux)
        _flux = numpy.ma.atleast_2d(_flux)
        if _flux.ndim > 2:
            raise ValueError('Flux must be 1 or 2D.')
        if _flux.shape[1] != len(_wave):
            raise ValueError('Flux has different spectral channels than the wavelength vector.')

        _ferr = ferr if isinstance(ferr, numpy.ma.MaskedArray) else numpy.ma.MaskedArray(ferr)
        _ferr = numpy.ma.atleast_2d(_ferr)
        if _ferr.shape != _flux.shape:
            raise ValueError('Flux and errors must have the same shape.')

        _mask = numpy.zeros_like(_flux, dtype=bool) if mask is None \
                        else numpy.atleast_2d(mask).astype(bool)
        if _mask.shape != _flux.shape:
            raise ValueError('Flux and mask must have the same shape.')
        _flux.mask |= _mask
        _ferr.mask |= _mask

        _model_flux = model_flux if isinstance(model_flux, numpy.ma.MaskedArray) \
                        else numpy.ma.MaskedArray(model_flux)
        _model_flux = numpy.ma.atleast_2d(_model_flux)
        if _flux.shape != _model_flux.shape:
            raise ValueError('Flux and model must have the same shape.')

        _model_mask = numpy.zeros_like(_flux, dtype=bool) if model_mask is None \
                        else numpy.atleast_2d(model_mask).astype(bool)
        if _model_mask.shape != _flux.shape:
            raise ValueError('Flux and model mask must have the same shape.')
        _model_flux.mask |= _model_mask

        # Spectra for figures of merit
        resid = numpy.square(_flux-_model_flux)
        fresid = numpy.square(numpy.ma.divide(_flux-_model_flux, _model_flux))
        chisqr = numpy.square(numpy.ma.divide(_flux-_model_flux, _ferr))
        spec_mask = resid.mask | fresid.mask | chisqr.mask
        resid[spec_mask] = numpy.ma.masked
        fresid[spec_mask] = numpy.ma.masked
        chisqr[spec_mask] = numpy.ma.masked

        # Get the pixel at which to center the metric calculations
        eml_mask = numpy.zeros(model_eml_par['MASK'].shape, dtype=bool) if bitmask is None else\
                        bitmask.flagged(model_eml_par['MASK'],
                                        flag=['INSUFFICIENT_DATA', 'FIT_FAILED', 'UNDEFINED_COVAR',
                                              'NEAR_BOUND'])
        z = numpy.ma.MaskedArray(model_eml_par['KIN'][:,:,0]/astropy.constants.c.to('km/s').value,
                                 mask=eml_mask)
        sample_wave = emission_lines['restwave'][None,:]*(1+z)
        interp = interp1d(wave, numpy.arange(wave.size), bounds_error=False, fill_value=-1,
                          assume_sorted=True)
        model_eml_par['LINE_PIXC'] = numpy.around(interp(sample_wave.data)).astype(int)
        model_eml_par['LINE_PIXC'][sample_wave.mask] = -1
        eml_mask = (model_eml_par['LINE_PIXC'] < 0) | (model_eml_par['LINE_PIXC'] >= wave.size)

        # Get the fitted line amplitude
        sigma_ang = model_eml_par['KIN'][:,:,1]*sample_wave/astropy.constants.c.to('km/s').value
        sigma_ang = numpy.ma.MaskedArray(sigma_ang, mask=numpy.invert(sigma_ang > 0))
        model_eml_par['AMP'] = numpy.ma.divide(model_eml_par['FLUX'], sigma_ang).filled(0.0) \
                                    / numpy.sqrt(2*numpy.pi)

        # Shift the bands to the appropriate redshift
        nspec = model_eml_par.size
        if numpy.any(z.mask) and fill_redshift:
            z_per_spec = numpy.ma.median(z, axis=1)
            for i in range(nspec):
                z[i,z.mask[i,:]] = z_per_spec[i]
        _bluebands = emission_lines['blueside'][None,:,:]*(1.0+z.data[:,:,None])
        _redbands = emission_lines['redside'][None,:,:]*(1.0+z.data[:,:,None])

        # Get the mean noise in the sidebands to either side of each
        # emission line
        noise = numpy.zeros_like(model_eml_par['AMP'], dtype=float)
        for i in range(nspec):
            _bluenoise = passband_median(_wave, _ferr[i,:], passband=_bluebands[i,:,:])
            _bluenoise = numpy.ma.MaskedArray(_bluenoise, mask=numpy.invert(_bluenoise > 0))
            _rednoise = passband_median(_wave, _ferr[i,:], passband=_redbands[i,:,:])
            _rednoise = numpy.ma.MaskedArray(_rednoise, mask=numpy.invert(_rednoise > 0))
            noise[i,:] = ((_bluenoise + _rednoise)/2.).filled(0.0)
        model_eml_par['ANR'] = numpy.ma.divide(model_eml_par['AMP'], noise).filled(0.0)

        neml = len(emission_lines)
        for i in range(neml):
            print('Getting fit metrics for line: {0}/{1}'.format(i+1, neml), end='\r')
            start = model_eml_par['LINE_PIXC'][:,i] - window//2
            end = start + window
            m = numpy.zeros(flux.shape, dtype=float)
            for j in range(nspec):
                if eml_mask[j,i]:
                    continue
                m[j,start[j]:end[j]] = 1.
            m[spec_mask] = 0.

            model_eml_par['LINE_NSTAT'][:,i] = numpy.sum(m,axis=1)
            model_eml_par['LINE_RMS'][:,i] = numpy.ma.sqrt(numpy.ma.mean(m*resid,axis=1)
                                                           ).filled(0.0)
            model_eml_par['LINE_FRMS'][:,i] = numpy.ma.sqrt(numpy.ma.mean(m*fresid,axis=1)
                                                            ).filled(0.0)
            model_eml_par['LINE_CHI2'][:,i] = numpy.ma.sum(m*chisqr,axis=1).filled(0.0)

        print('Getting fit metrics for line: {0}/{0}'.format(neml))
        return model_eml_par


